#!/usr/bin/env python3
'''
wbutil/math/tree.py

Classes for creating easy to use decision trees.

Will Badart <wbadart@live.com>
created: FEB 2018
'''

from csv import reader
from functools import lru_cache, partial
from math import log2 as lg
from operator import eq, ge, lt, itemgetter
from os import PathLike
from typing import Generator, Iterable, Tuple, Union
from uuid import uuid4

from graphviz import Digraph
from wbutil import compose

__all__ = [
    'DecisionTree',
]

_FeatureListT = Iterable[Union[str, Tuple[str, bool]]]
_PredicateT = compose[bool]


class DecisionTree(object):
    '''Represents a decision tree which branches based on information gain.'''

    # ==========
    # Public API
    # ==========

    def __init__(
            self,
            feature_names: _FeatureListT,
            *data: Iterable,
            test: _PredicateT=None,
            root_majority: str=None) -> None:
        '''
        Initialize the model.

        features:
            Ordered iterable of the feature names as strings. If an element
            of `features' is a 2-tuple, its first item is treated as the
            feature name and the second as a flag for continuous features (data
            in this column will be cast to float). The label/ target feature
            should occupy the last column.
        data:
            Each item in data should be an iterable of equal width to the
            feature list, with the last item holding the instance's label.
        test:
            Predicate to control which instances the tree claims for
            classification. Should not be set by user code.
        root_majority:
            Tracks which label was most common at the root node of the tree.
            Used to break ties at the leaves. If it is None (which it should be
            at the root node) it is computed to be the most frequent label.
        '''

        def ispair(feature):
            return type(feature) is tuple and len(feature) == 2

        features = tuple(feature_names)
        self._raw_features = features
        self._test = test

        # Find the features marked as continuous
        self._continuous = frozenset(
            f[0] for f in features if ispair(f) and f[1] is True)

        # Flatten the feature list (remove continuous labels)
        self._features = tuple(f if not ispair(f) else f[0] for f in features)
        continuous_idx = frozenset(map(self._features.index, self._continuous))

        # Continuous data points should be real numbers
        self._data = [tuple(float(e) if i in continuous_idx else e
                            for i, e in enumerate(tup)) for tup in data]
        self._labels = frozenset(tup[-1] for tup in self._data)
        self._root_majority = root_majority or max(self._labels,
                                                   key=self._count_label)

    def classify(self, instance: tuple) -> str:
        '''
        Predict the label of `instance'. `instance' can be a tuple of length
        equal to or one less than the width of the training data. The former
        case means `instance' has its label in the last position, in which case
        it is ignored.
        If the tree has branches and none accepts the instance, then either the
        instance or the branching is malformed, and a ValueError is raised.
        '''
        for branch in self._branches:
            if branch._test(instance):
                return branch.classify(instance)
        if self._has_branches:
            raise ValueError('no branch accepted %r' % (instance,))
        return self.prediction

    def show(self, dot: Digraph=None, is_root: bool=True) -> Digraph:
        '''
        Give a graphviz visualization of the tree. The two parameters are for
        recursive calls and should be ignored by user code.
        '''
        dot = dot or Digraph()
        key = str(uuid4())
        label = self._best_feature if self._has_branches else self.prediction
        dot.node(key, label=label)
        for branch in self._branches:
            dot, child_key = branch.show(dot, is_root=False)
            test = branch._test.funcs[1]
            dot.edge(key, child_key, label='%s %s' % (
                getattr(test, 'func').__name__, getattr(test, 'args')[0]))
        return dot if is_root else (dot, key)

    @property
    def prediction(self) -> str:
        '''
        The label this node would give if it were a leaf. Examines the training
        data and picks the most popular label. If multiple classes have the
        same frequency, returns the root majority.
        '''
        most_popular = [
            (label, self._count_label(label)) for label in (self._labels)]
        return (most_popular[0][0] if len(self._labels) == 1
                or most_popular[0][1] > most_popular[1][1]
                else self._root_majority)

    @classmethod
    def from_csv(
            cls,
            path: Union[str, PathLike],
            delimiter: str=',',
            features: _FeatureListT=None) -> 'DecisionTree':
        '''
        Load features and training data straight from csv file. `path'
        specifies the location of the file, `delimiter' is the cell separator
        passed to csv.reader, and `features' is a list of feature labels (as
        described in __init__). If `features' is None, the first row of the
        file will be interpreted as the feature labels and they will all
        be discrete features.
        '''
        with open(path) as fs:
            r = reader(fs, delimiter=delimiter)
            features = features or tuple(next(r))
            return cls(features, *r)

    # ====================
    # Entropy calculations
    # ====================

    def _information_gain(self, *tests: _PredicateT) -> float:
        '''
        Compute the information gain of the feature splitting encoded in
        `tests'.
        IG(Y|X=x) = H(Y) - H(Y|X=x)
        '''
        return self._entropy - self._conditional_entropy(*tests)

    def _conditional_entropy(self, *tests: _PredicateT) -> float:
        '''
        Compute the conditional entropy of class Y given feature X. X is
        encoded into `tests', which is a series of predicates which partition
        the data into mutually exclusive subsets.
        H(Y|X) = sum[x](p(x) * H(Y|X=x))
        '''
        entropy = 0.
        for test in tests:
            # `count' is the number of instances that pass the test (i.e. that
            # belong to the branch)
            count = self._count(test)
            # `frequency' is the number of those passing instances that have
            # each label
            frequencies = list(
                sum(1 for t in self._data if test(t) and t[-1] == label)
                for label in self._labels)
            # `total' represents the entropy contributed by this branch
            try:
                total = -sum(
                    (freq / count) * lg(freq / count) for freq in frequencies)
            except (ValueError, ZeroDivisionError):
                continue
            entropy += (count / len(self._data)) * total
        return entropy

    @property  # type: ignore
    @lru_cache(maxsize=1)
    def _entropy(self) -> float:
        '''
        Gives the unconditional entropy of the label class.
        H(Y) = -sum[i:=1->m](p_i log(p_i)) where p_i = P(Y = y_i)
        '''
        probs = (c / len(self._data)
                 for c in map(self._count_label, self._labels))
        return -sum(p * lg(p) for p in probs)

    SPLIT_CRITERION = _information_gain

    # =============================
    # Helper methods and properties
    # =============================

    def _best_split(self, feature: str) -> float:
        '''
        Compute the best pivot for continuous `feature'. Scans the observed
        feature values and computes the one with the greatest split metric.
        '''
        if feature not in self._continuous:
            raise TypeError('expecting a continuous feature')
        feature_index = self._features.index(feature)
        return max(
            frozenset(t[feature_index] for t in self._data),
            key=lambda value: self.SPLIT_CRITERION(
                *self._gen_tests(feature, value)))

    def _count(self, test: _PredicateT) -> int:
        '''
        Count the number of instances which pass predicate `test' (which should
        be a function of a data tuple).
        '''
        return sum(1 for t in self._data if test(t))

    def _count_label(self, label: str) -> int:
        '''
        Count the number of tuples with label `label'. Leverages
        DecisionTree._count by constructing the predicate which determines if a
        tuple has the argument label and calling it for the summation.
        '''
        has_label: compose[bool] = compose(itemgetter(-1), partial(eq, label))
        return self._count(has_label)

    def _gen_tests(self, feature: str, split: float=None) -> Generator:
        '''
        Generate the tests which branch `feature'. Yields a series of
        predicates that partition this node's data into mutually exclusive
        subsets (i.e. for each instance, exactly one of the yielded predicates
        will be true).
        '''
        feature_index = self._features.index(feature)
        if feature in self._continuous:
            split = split if split is not None else self._best_split(feature)
            yield compose(itemgetter(feature_index), partial(lt, split))
            yield compose(itemgetter(feature_index), partial(ge, split))
        else:
            for val in frozenset(t[feature_index] for t in self._data):
                yield compose(itemgetter(feature_index), partial(eq, val))

    @property
    def _best_feature(self) -> str:
        '''
        The feature which gives the best SPLIT_CRITERION metric. Considers both
        discrete and continuous features (only excludes label class.).
        '''
        return max(self._features[:-1],
                   key=lambda f: self.SPLIT_CRITERION(*self._gen_tests(f)))

    @property
    def _branches(self) -> Iterable['DecisionTree']:
        '''
        Finds the optimal branching of this tree and returns the computed
        subtrees.
        '''
        if not self._has_branches:
            return []
        for test in self._gen_tests(self._best_feature):
            yield type(self)(
                self._raw_features,
                *filter(test, self._data),
                test=test,
                root_majority=self._root_majority)

    @property
    def _has_branches(self) -> bool:
        '''
        Report whether the node will branch or not (i.e. whether terminal
        conditions have been met yet).
        '''
        return len(self._data) > 10 and (
            self.SPLIT_CRITERION(
                *self._gen_tests(self._best_feature)) > 0.00001)



#!/usr/bin/env python

'''
wbutil/misc.py

Miscellaneous utils with no real home elsewhere.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

import signal
from contextlib import ContextDecorator
from functools import wraps
from itertools import count
from numbers import Real
from random import random
from time import sleep, time
from typing import (
    Any, Callable, Generator, Iterable, Optional, Tuple, Type, Union)

__all__ = [
    'pair',
    'randreal',
    'retry',
    'signal_handler',
    'slow',
    'timeout',
    'uniq',
]


def pair(seq: list, e: Any) -> list:
    '''
    For reducing a list to a list of pairs of adjacent elements.

    >>> reduce(pair, range(6), [tuple()])
    [(0, 1), (2, 3), (4, 5)]
    '''
    if len(seq[-1]) == 2:
        seq.append((e,))
    else:
        seq[-1] += (e,)
    return seq


def randreal(min_: Real, max_: Real) -> Real:
    '''
    Generate a pseudo-random number in [min_, max_].

    >>> randreal(10.5, 15.5)
    12.789728304503505
    '''
    return random() * (max_ - min_) + min_


def retry(
        times: Optional[int]=None,
        wait: int=0,
        exceptions: Union[Type[Exception], Tuple[Type[Exception]]]=Exception,
        default: Any=None) -> Callable:
    '''
    Returns a version of `f' that will automatically retry `times' times if the
    invokation raises any of the given exceptions. Will wait `wait' seconds to
    try again. If `times' is None, will retry indefinitely. Otherwise, when
    retry has reached that many of attempts, it will return `default' or raise
    ValueError if default is None.

    To be used as a decorator, parameters should be given as keyword arguments.

    >>> import requests as r
    >>> fetch = retry(r.get, times=3, exceptions=ConnectionError)
    >>> fetch(URL)
    <Response [200]>
    '''
    def _impl(f):
        @wraps(f)
        def _wrapper(*args, **kwargs):
            attempts = count() if times is None else range(times)
            first_pass = True
            for _ in attempts:
                if not first_pass:
                    sleep(wait)
                first_pass = False
                try:
                    return f(*args, **kwargs)
                except exceptions:
                    pass
            if default is not None:
                return default
            else:
                raise ValueError(
                    'Function application failed after %d attempts' % times)
        return _wrapper
    return _impl


def signal_handler(signum: int) -> Callable:
    '''
    Registers a function as the handler of signal `signum'.

    >>> import signal as sig; from os import kill, getpid
    >>> @signal_handler(sig.SIGUSR1)
    ... def f(signum, frame):
    ...     print('Got SIGUSR1')
    ...
    >>> f(None, None)
    Got SIGUSR1
    >>> kill(getpid(), sig.SIGUSR1)
    Got SIGUSR1
    '''
    def _impl(func: Callable[[int, Any], None]) -> Callable[[int, Any], None]:
        signal.signal(signum, func)
        return func
    return _impl


class timeout(ContextDecorator):
    '''
    Try to execute a block of code, but preempt (raise TimeoutError) if it
    takes too long.

    >>> from time import sleep
    >>> @timeout(1)
    ... def test():
    ...     sleep(5)
    ...     print('Done sleeping')
    ...
    >>> test()
    Traceback (most recent call last):
      File "<string>", line 1, in <module>
    TimeoutError: Time allotment of 1 second expired
    >>> with timeout(10):
    ...     sleep(5)
    ...     print('Waited 5 seconds')
    ...
    Waited 5 seconds
    '''

    def __init__(self, duration: int=5) -> None:
        self.duration = duration
        signal.signal(signal.SIGALRM, self._raisetimeout)

    def __enter__(self) -> 'timeout':
        '''Start the block and the timer.'''
        signal.alarm(self.duration)
        return self

    def __exit__(
            self,
            exception_t=Exception,
            exception=None,
            traceback=None) -> None:
        '''Cancel the alarm when block finishes.'''
        signal.alarm(0)

    def _raisetimeout(self, signum: int, frame: Any) -> None:
        '''Trap the alarm signal.'''
        raise TimeoutError(
            'Time allotment of %d second%s expired'
            % (self.duration, 's' if self.duration != 1 else ''))


class slow(object):
    '''
    Slows the execution of a function. Like a throttle except all calls are
    eventually executed (in sequence).

    >>> from wbutil import timeout
    >>> slow_print = slow(print, 1)
    >>> with timeout(5):
    ...     for i in range(10):
    ...         slow_print(i)
    ...
    0
    1
    2
    3
    4
    Traceback (most recent call last):
      File "<stdin>", line 3, in <module>
      File "wbutil/misc.py", line 150, in __call__
      File "wbutil/misc.py", line 128, in _raisetimeout
    TimeoutError: Time allotment of 5 seconds expired
    '''

    def __init__(self, wait: Union[int, float]) -> None:
        self.wait = wait
        self.last_call = 0.

    def __call__(self, func: Callable) -> Any:
        '''
        '''
        @wraps(func)
        def _impl(*args, **kwargs):
            time_elapsed = time() - self.last_call
            if time_elapsed < self.wait:
                sleep(self.wait - time_elapsed)
            self.last_call = time()
            return func(*args, **kwargs)
        return _impl


def uniq(a: Iterable) -> Generator:
    '''
    Generate the argument iterator with the duplicate members removed.

    >>> list(uniq([1, 2, 2, 3]))
    [1, 2, 3]
    '''
    seen: set = set()
    for e in a:
        if e not in seen:
            seen.add(e)
            yield e
        else:
            continue



#!/usr/bin/env python

'''
wbutil/math/stat.py

Various utilities for descriptive statistics.

Will Badart <wbadart@live.com>
created: FEB 2018
'''

from collections import Counter
from numbers import Real
from typing import Tuple, Sequence

__all__ = [
    'ConfusionMatrix',
    'mean',
    'median',
    'variance',
    'std',
]


class ConfusionMatrix(object):
    '''
    Track the performance of a binary classifier. Report performance measures
    such as accuracy, recall, precision, and balanced F1 score.

    >>> m = ConfusionMatrix('win', 'lose')
    >>> m.updates(('win', 'win'), ('win', 'lose'))
    >>> m.accuracy
    0.5
    '''

    def __init__(self, positive_label: str, negative_label: str) -> None:
        self._pos = positive_label
        self._neg = negative_label
        self._c: Counter = Counter(TP=0, FP=0, TN=0, FN=0)

    def update(self, actual_label: str, predicted_label: str) -> None:
        '''
        Report a prediction and the corresponding true label. If either of the
        argument labels do not correspond to either the positive or negative
        labels of this matrix, a ValueError is raised.
        '''
        if actual_label not in self._labels:
            raise ValueError('label %r is not a known label' % actual_label)
        if predicted_label not in self._labels:
            raise ValueError(
                'prediction %r is not a known label' % predicted_label)

        if predicted_label == self._pos:
            if predicted_label == actual_label:
                self._c['TP'] += 1
            else:
                self._c['FP'] += 1

        elif predicted_label == self._neg:
            if predicted_label == actual_label:
                self._c['TN'] += 1
            else:
                self._c['FN'] += 1

    def updates(self, *results: Tuple[str, str]) -> None:
        '''
        Process a list of classifications. Each `result' should be a 2-tuple
        with the actual label on the left and the predicted label on the right.
        '''
        for actual, predicted in results:
            self.update(actual, predicted)

    @property
    def accuracy(self) -> float:
        '''
        Report the accuracy of the model (proportion of predictions which we
        correct.
        (TP + TN) / |observations|
        '''
        return (self.TP + self.TN) / len(self)

    @property
    def error(self) -> float:
        '''
        Report the model's error rate (another way to frame accuracy). Gives
        1 - accuracy or:
        (FP + FN) / |observations|
        '''
        return (self.FP + self.FN) / len(self)

    @property
    def sensitivity(self) -> float:
        '''
        True positive recognition rate (for data sets with heavy class
        imbalance).
        TP / P aka TP / (TP + FN)
        '''
        return self.TP / (self.TP + self.FN)

    @property
    def specificity(self) -> float:
        '''
        True negative recognition rate
        TN / N aka TN / (TN + FP)
        '''
        return self.TN / (self.TN + self.FP)

    @property
    def precision(self) -> float:
        '''
        Report the precision of the model. Signals how exact the model is by
        calculating the proportion of positive labels which corresponded to
        actual positive instances.
        TP / P' where P' is the total number of positive predictions, TP + FP
        '''
        return self.TP / (self.TP + self.FP)

    @property
    def recall(self) -> float:
        '''
        Proportion of positive instances labeled as such by the model.
        TP / P (same as sensitivity)
        '''
        return self.TP / (self.TP + self.FN)

    @property
    def f1(self) -> float:
        '''
        Gives the balanced F measure for the model.
        2PR / (P + R) where P and R are precision and recall respectively.
        '''
        return 2 * self.precision * self.recall \
            / (self.precision + self.recall)

    @property
    def TP(self) -> int:
        '''
        Count of true positives (actual and prediction both positive) observed.
        '''
        return self._c['TP']

    @property
    def FP(self) -> int:
        '''
        Count of false positives (predicted positive for negative instance)
        observed.
        '''
        return self._c['FP']

    @property
    def TN(self) -> int:
        '''
        Count of true negatives (actual and prediction both negative) observed.
        '''
        return self._c['TN']

    @property
    def FN(self) -> int:
        '''
        Count of false negatives (predicted false for a positive instance)
        observed.
        '''
        return self._c['FN']

    @property
    def _labels(self) -> set:
        return {self._pos, self._neg}

    def __len__(self) -> int:
        '''Corresponds to the number of predictions observed.'''
        return sum(self._c.values())

    def __str__(self) -> str:
        '''Gives a nice, relatively pretty-printed table of results.'''
        return (
            f'Actual \\ Predicted | {self._pos} | self._neg'
            f'    {self._pos} | {self.TP} | {self.FN}'
            f'    {self._neg} | {self.FP} | {self.TN}')


def mean(a: Sequence[Real]) -> float:
    '''
    Return the mean of a list of numbers.

    >>> mean([1, 2, 3, 4])
    2.5
    '''
    return sum(a) / len(a)


def median(a: Sequence[Real]) -> float:
    '''
    Return the median of a list of numbers.

    >>> median([4, 6, 1])
    4
    >>> median([1, 1, 2, 3])
    1.5
    '''
    if _isodd(len(a)):
        return float(sorted(a)[len(a) // 2])
    else:
        return mean(sorted(a)[len(a) // 2 - 1:len(a) // 2 + 1])


def variance(a: Sequence[Real]) -> float:
    '''
    Get the variance (average squared distance to the mean) of a number list.

    >>> variance([1, 2, 3])
    0.6666666666666666
    >>> variance([1, 1, 1, 1])
    1.0
    '''
    avg = mean(a)
    return mean([(avg - e) ** 2 for e in a])


def std(a: Sequence[Real]) -> float:
    '''
    Gives the standard deviation of a list of numbers (sqrt(variance)).
    '''
    return variance(a) ** 0.5


def _isodd(n):
    return n % 2 != 0



#!/usr/bin/env python

'''
wbutil/__init__.py

Main manifest of wbutil modules.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

from . import coroutine
from . import func
from . import fs
from . import math
from . import misc
from . import pipeline

from .coroutine import *
from .fs import *
from .func import *
from .math import *
from .misc import *
from .pipeline import *



#!/usr/bin/env python

'''
wbutil/func.py

Functional-programming-oriented utilities.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

import re
from functools import partial, reduce, wraps
from inspect import signature
from itertools import chain
from typing import Any, Callable, Generic, Iterable, TypeVar, Union

__all__ = [
    'autocurry',
    'compose',
    'partialright',
    'starcompose',
    'cmap',
    'cfilter',
    'creduce',
    'lmap',
    'lfilter',
    'identity',
    'pred_negate',
]

_ComposeRetT = TypeVar('_ComposeRetT')
_CurryReturnT = TypeVar('_CurryReturnT')
_NAryFunc = Callable[[Iterable[Any]], Iterable[Any]]
_UnaryFunc = Callable[[Any], Any]
_T = TypeVar('_T')


class autocurry(Generic[_CurryReturnT]):
    '''
    Wraps a callable which it autocurries. Calling the function with fewer than
    the total number of parameters yields a partial application, and fulfilling
    the parameter list yields a complete application. Avoid complications by
    not applying this to functions of variable parameters.

    >>> from operator import add
    >>> a = autocurry(add)
    >>> a(3)(4)
    7
    '''

    def __init__(
            self,
            func: Callable[..., _CurryReturnT],
            *args,
            **kwargs) -> None:
        self.func = func
        try:
            self.nparams = len(signature(func).parameters)
        except ValueError:
            func = getattr(func, 'func', func)  # if func is a partial
            prototype = re.search('^[\w_]+(\(.*\))', func.__doc__).groups()[0]
            prototype = prototype[1:-1]
            self.nparams = len(prototype.split(','))

        self.args = args
        self.keywords = kwargs

    def __call__(self, *args, **kwargs) -> Union['autocurry', _CurryReturnT]:
        '''
        Partially apply args and kwargs to the wrapped function (full
        application if all parameters have been filled.
        '''
        pos_args = chain(self.args, args)
        keywords = dict(**self.keywords, **kwargs)
        if len(self.args + args) == self.nparams:
            return self.func(*pos_args, **keywords)
        else:
            return type(self)(self.func, *pos_args, **keywords)


class compose(Generic[_ComposeRetT]):
    '''
    Returns a function which calls the argument functions in order from left
    to right (first to last). Undefined behavior for non-unary functions.

    >>> from functools import partial; from operator import mul
    >>> times2tostr = compse(partial(mul, 2), str)
    >>> times2tostr(10)
    '20'
    '''

    def __init__(self, *funcs: _UnaryFunc) -> None:
        self.funcs = funcs

    def __call__(self, arg: Any) -> _ComposeRetT:
        '''Invoke the composed pipeline with the specified argument.'''
        return reduce(lambda acc, f: f(acc), self.funcs, arg)

    def __reversed__(self) -> 'compose':
        '''
        Gives a composition with the calling order reversed. Allows compose to
        emulate traditional compositional ordering.
        '''
        funcs = tuple(reversed(self.funcs))
        return type(self)(*funcs)


class starcompose(compose):
    '''
    Returns a function of *args that applies the argument functions in
    order from left to right (first to last). All supplied functions must
    return an iterable and accept *args.

    >>> switch = lambda a, b: (b, a)
    >>> tup2str = lambda a, b: (str(a), str(b))
    >>> switchandstr = starcompose(switch, tup2str)
    >>> switchandstr(1, 2)
    ('2', '1')
    '''

    # Overridden to annotate n-ary function support
    def __init__(self, *funcs: _NAryFunc) -> None:
        super().__init__(*funcs)

    def __call__(self, *args: Any) -> Any:
        '''Invoke the composed pipeline with unpacking.'''
        return reduce(lambda acc, f: f(*acc), self.funcs, args)


class partialright(partial):
    '''
    Returns a new function with arguments partially applied from right to left.

    >>> data = [1, 2, 3, 4]
    >>> process_data = partialright(map, data)
    >>> stringified = process_data(str)
    >>> stringified
    ['1', '2', '3', '4']
    '''

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        '''Call self as a function.'''
        pos_args = chain(args, reversed(self.args))
        keywords = self.keywords.copy()
        keywords.update(kwargs)
        return self.func(*pos_args, **keywords)


def cmap(iteratee: Callable[[Any], Any]) -> partial:
    '''
    Curried map. Sugar for partial(map, iteratee).

    >>> from string import digits
    >>> transform = cmap(int)
    >>> list(transform(digits))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    '''
    return partial(map, iteratee)


def cfilter(predicate: Callable[[Any], bool]) -> partial:
    '''
    Curried filter. Sugar for partial(filter, predicate).

    >>> is1or2 = cfilter({1, 2}.__contains__)
    >>> list(is1or2(range(10)))
    [1, 2]
    '''
    return partial(filter, predicate)


def creduce(iteratee: Callable[[Any], Any]) -> partial:
    '''
    Curried reduce. Allows sequence to be supplied later.

    >>> from operator import add
    >>> mysum = creduce(add)
    >>> mysum(range(10)) == sum(range(10))
    True
    >>> from collections import defaultdict
    >>> def byfirst(obj, e):
    ...     obj[e[0]].append(e)
    ...     return obj
    ...
    >>> groupbyfirst = creduce(byfirst)
    >>> groupbyfirst(db.getNames(), defaultdict(list))
    defaultdict(<class 'list'>, {'A': ['Alice', 'Adam'], 'B': ['Bob']})
    '''
    return partial(reduce, iteratee)


def lmap(iteratee: Callable[[Any], Any], iterable: Iterable) -> list:
    '''
    Performs a mapping of `iteratee' over `iterable' and immediately
    serializes result into a list.

    >>> lmap(str, range(10))
    ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
    '''
    return list(map(iteratee, iterable))


def lfilter(predicate: Callable[[Any], bool], iterable: Iterable) -> list:
    '''
    Filters `iterable' on `predicate', immediately serializing results into a
    list.

    >>> odd = lambda x: x % 2 == 1
    >>> lfilter(odd, range(10))
    [1, 3, 5, 7, 9]
    '''
    return list(filter(predicate, iterable))


def identity(e: _T) -> _T:
    '''
    Returns the argument untouched. Useful default iteratee.

    >>> list(map(identity, range(10)))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    '''
    return e


def pred_negate(predicate: Callable[..., bool]) -> Callable[..., bool]:
    '''
    Gives a function which returns the inverse of the argument predicate.

    >>> isodd = lambda x: x % 2 == 1
    >>> iseven = pred_negate(isodd)
    >>> iseven(10)
    True
    '''
    @wraps(predicate)
    def _impl(*args, **kwargs):
        return not predicate(*args, **kwargs)
    return _impl



#!/usr/bin/env python

'''
wbutil/math/naivebayes.py

Naïve Bayes classifier.

Will Badart <netid:wbadart>
created: FEB 2018
'''

from collections import Counter, defaultdict
from csv import reader
from pprint import pformat
from typing import Sequence

__all__ = [
    'NaiveBayes',
]


class NaiveBayes(object):
    '''
    NaiveBayes objects represent a classifier trained for the given context.
    '''

    def __init__(
            self,
            features: Sequence[str],
            *data: tuple,
            label_index=-1,
            domains=None) -> None:
        '''
        Construct a new naive bayes model with features `features' and training
        instances `data'.

        Specify the column of the label with `label_index' (negative numbers
        are interpreted as index from the right, e.g. label_index = -2 if the
        label is in the second to last column.

        `domains' is a list of types that correspond 1-1 with the features.
        Each domain can be any callable to be applied to the cells of that
        column. Defaults to identity.
        '''
        self._features = features
        self._data = data
        self._label_index = (
            label_index if label_index >= 0 else len(features) + label_index)
        self._label = features[self._label_index]
        self._domains = domains or (_identity,) * len(features)

        # self._counts tracks, for each class for each feature, the number of
        # occurrences of each label
        self._counts: dict = {
            feature: defaultdict(Counter) for feature in features}

        for instance in data:
            instance = tuple(
                self._domains[i](val) for i, val in enumerate(instance))
            for feat, val in self._genfeats(instance):
                self._counts[feat][val][instance[self._label_index]] += 1
            # indexing None moves us from the defaultdict into the Counter for
            # the label counts
            self._counts[self._label][None][instance[self._label_index]] += 1

    def predict(self, instance):
        '''
        Predict the label of the given unlabeled test instance.

            Y = argmax[yj in Y] P(yj) * PRODUCT[P(xi|yj) for xi in X]
        '''
        maxlabel, maxprob = None, 0.0
        for label, count in self._counts[self._label][None].items():

            # prior probability
            labelprob = count / len(self._data)

            # probability of feature=value given current label
            for feat, val in self._genfeats(instance):
                labelprob *= self.probability(feat, val, label)

            if labelprob >= maxprob:
                maxprob = labelprob
                maxlabel = label

        return maxlabel, maxprob

    def probability(self, feature, value, class_):
        '''Read class counts to determine the probability of x_i give y_j.'''
        count = self._counts[feature][value][class_]
        # the frequency of label `class_' for all values of `feature'
        total = sum(ct[class_] for ct in self._counts[feature].values())
        return count / total

    def prediction_report(self, instance):
        '''
        Predict the label of the given instance and return a pretty string with
        the result.
        '''
        prediction, probability = self.predict(instance)
        return (
            'Prediction:\n'
            '    Tuple:       %s\n'
            '    Label:       "%s"\n'
            '    Probability: %f'
            % (instance, prediction, probability))

    @classmethod
    def from_csv(cls, path: str, label_index: int=-1) -> 'NaiveBayes':
        '''
        Generate a Naive Bayes model from a csv file, where the first row
        contains the feature names. label_index follows same convention as
        __init__.
        '''
        with open(path) as fs:
            return cls(*map(tuple, reader(fs)), label_index=label_index)

    def _skiplabel(self, tup):
        '''Iterate over a tuple, skipping the label column.'''
        return (e for i, e in enumerate(tup) if i != self._label_index)

    def _genfeats(self, instance):
        '''
        Iterate over a tuple, including feature name and skipping label column.
        '''
        return self._skiplabel(zip(self._features, instance))

    def __repr__(self):
        return '{cls}({features!r},\n{data},\nlabel_index={idx})'.format(
            cls=type(self).__name__,
            features=self._features,
            data=pformat(self._data),
            idx=self._label_index)

    def __str__(self):
        features = (
            f if i != self._label_index else '*' + f
            for i, f in enumerate(self._features))
        return '{cls}({features})'.format(
            cls=type(self).__name__, features=', '.join(features))


def _identity(a):
    '''Returns the argument.'''
    return a



#!/usr/bin/env python

'''
wbutil/fs.py

Filesystem utilities.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

from contextlib import closing
from json import dumps, loads
from os import PathLike
from os.path import exists
from pickle import dump, load
from typing import Any, Callable, TextIO, Union

__all__ = [
    'saveobj',
    'tryopen',
]


def saveobj(constructor: Callable[[], Any], path: str) -> Any:
    '''
    Construct an object and save it to the disk for later use. Useful for
    objects with expensive constructors such as ones that include Web requests
    since they're only run if the storage file doesn't already exist.

    >>> from functools import partial; from requests import get
    >>> req = objstorage(partial(get, 'http://example.com'), 'myreq.pickle')
    >>> req
    <Response [200]>
    '''
    if not exists(path):
        obj = constructor()
        with open(path, 'wb') as fs:
            dump(obj, fs)
    else:
        with open(path, 'rb') as fs:
            obj = load(fs)
    return obj


def tryopen(
        path: str,
        process: Callable[[TextIO], Any]=lambda fs: fs.read(),
        default: Any=None) -> Any:
    '''
    Attempt to open the file at `path`, process it, and close it. If file
    cannot be loaded, the default value is returned. Does not swallow errors
    from the processing function.

    >>> from json import load
    >>> data = try_open('data.json', process=load, default={})
    >>> data
    {'my': 'data'}
    >>> description = try_open('doesnt_exist.txt', default='No description')
    >>> description
    'No description'
    '''
    try:
        fs = open(path)
    except OSError as e:
        if default is not None:
            return default
        else:
            msg = ('an error occurred opening {!r}. '
                   'Provide a non-None argument `default\' to suppress')
            raise OSError(msg.format(path)) from e
    with closing(fs):
        return process(fs)


class PersistentDict(dict):
    '''
    A dict subclass, objects of which are directly linked to a file. That is to
    say, it provides methods for loading and saving self to and from a file.

    Provides a context manager interface so that data will be saved even if an
    exception is thrown in processing.

    Serialization protocol is JSON by default, but you can use any encoding
    sends a dictionary to a string (e.g. YAML, picke).
    '''

    def __init__(
            self,
            *args: Any,
            path: Union[str, PathLike]=None,
            encode: Callable[[dict], str]=dumps,
            usebytes: bool=False) -> None:
        '''
        Initialize like an ordinary dictionary, but provide the keyword-only
        arguments `path' (the file liked to this object) and `encode' (the
        function which serializes this object into a string to write to disk,
        defaults to `json.dumps').

        Set the `usebytes' flag in the constructor if your protocol expects
        bytes objects rather than strings.

        Assumes `path' is fresh. Use classmethod `from_file' to start with data
        loaded.
        '''
        if path is None:
            raise ValueError('cannot use PersistentDict without a save path')
        super().__init__(*args)
        self.path = path
        self.encode = encode
        self.mode = 'b' if usebytes else ''

    def __enter__(self) -> 'PersistentDict':
        return self

    def __exit__(
            self,
            exception_t=Exception,
            exception=None,
            traceback=None) -> None:
        '''
        Save the object when context is exited (throw away return value).
        '''
        self.save()

    def save(self) -> int:
        '''
        Write self's data to disk. Returns the result of `file.write' (total
        bytes written to disk).
        '''
        with open(self.path, 'w' + self.mode) as fs:
            return fs.write(self.encode(self))

    @classmethod
    def from_file(
            cls,
            path: Union[str, PathLike],
            encode: Callable[[dict], str]=dumps,
            decode: Callable[[str], dict]=loads,
            usebytes: bool=False) -> 'PersistentDict':
        '''
        Instantiate a PersistentDict from an existing data file. Arguments
        follow the same semantics as PersistentDict.__init__, and `decode'
        should be the inverse of `encode'.
        '''
        with open(path, 'rb' if usebytes else 'r') as fs:
            obj = decode(fs.read())
        return cls(obj, path=path, encode=encode, usebytes=usebytes)



#!/usr/bin/env python

'''
wbutil/math/__init__.py

Top-level manifest of the math utilities.

Will Badart <wbadart@live.com
created: FEB 2018
'''

from . import stat
from . import naivebayes

from .naivebayes import *
from .stat import *



#!/usr/bin/env python

'''
wbutil/pipeline.py

Helpers which wrap standard library threaded queue.

Will Badart <wbadart@live.com>
created: JAN 2018
'''

from collections import OrderedDict
from queue import Queue as StdQueue
from threading import Event, Lock, Thread
from typing import Callable, Dict, Hashable, Iterable, List, TypeVar
from uuid import uuid4

__all__ = [
    'Pipeline',
]

_FuncArgT = TypeVar('_FuncArgT')
_FuncRetT = TypeVar('_FuncRetT')


class Pipeline(object):
    '''
    Simplify the interface provided by standard library queue. Specifically
    tailored to processing lists of data.
    '''

    def __init__(
            self,
            func: Callable[[_FuncArgT], _FuncRetT],
            maxsize: int=0,
            nthreads: int=4) -> None:

        self._q: StdQueue = StdQueue(maxsize)
        self._func = func

        self._threads: List[Thread] = []
        for _ in range(nthreads):
            self._threads.append(Thread(target=self._worker))
        self._nthreads = nthreads

        self._results: OrderedDict[Hashable, _FuncRetT] = OrderedDict()
        self._result_lock = Lock()
        self._events: Dict[Hashable, Event] = {}

        # Track queue state; enforce one-time usage
        self._started = False
        self._stopped = False

    def _worker(self) -> None:
        while True:
            key, item, event = self._q.get()
            if key is None:
                self._q.task_done()
                break
            result = self._func(item)
            with self._result_lock:
                self._results[key] = result
            event.set()
            self._q.task_done()

    def __enter__(self) -> 'Pipeline':
        '''Start threads.'''
        if self._started:
            raise RuntimeError('Pipelines cannot be reused')
        self._started = True
        for thread in self._threads:
            thread.start()
        return self

    def __exit__(
            self,
            exception_t=Exception,
            exception=None,
            traceback=None) -> None:
        '''Stop and join the threads.'''
        for thread in self._threads:
            self._q.put((None, None, None))
        for thread in self._threads:
            thread.join()
        self._q.join()
        self._stopped = True

    def put(self, item: _FuncArgT, key: Hashable=None) -> Hashable:
        '''Add an item to apply func to.'''
        if self._stopped:
            raise RuntimeError('this Pipeline has already been stopped')
        if key is None and item is not None:
            key = str(uuid4())
        e = Event()
        self._events[key] = e
        self._q.put((key, item, e))
        return key

    def get(self, key: Hashable, timeout: int=None) -> _FuncRetT:
        '''Get the result from an individual item.'''
        self._events[key].wait(timeout)
        return self._results.get(key)

    def map(self, iterable: Iterable[_FuncArgT]) -> List[_FuncRetT]:
        '''
        Apply Pipeline.func to each item in the argument iterable.
        '''
        if not self._started:
            self.start()
        for i, item in enumerate(iterable):
            self.put(item, key=i)
        self.shutdown()
        return [self._results[j] for j in range(i + 1)]

    def apply(self, iterable: Iterable[_FuncArgT]) -> None:
        '''
        Procedure variant of map. Apply Pipeline.func over iterable, but do not
        aggregate and return procedure results.
        '''
        if not self._started:
            self.start()
        for e in iterable:
            self.put(e)
        self.shutdown()

    @property
    def func(self) -> Callable[[_FuncArgT], _FuncRetT]:
        '''Allow user code to inspect (though not modify) func.'''
        return self._func

    # Nice aliases for context functions, for use when `with' not appropriate
    start = __enter__
    shutdown = __exit__



#!/usr/bin/env python3

'''
wbutil/coroutine.py

Helpers for easy coroutine usage.

Will Badart <wbadart@live.com>
created: FEB 2018
'''

from functools import wraps
from pprint import pprint
from typing import Callable, Coroutine, Generator

__all__ = [
    'broadcast',
    'prime_coroutine',
]


def prime_coroutine(func: Callable) -> Callable:
    '''
    Takes a generator-based coroutine and starts it immediately after
    construction (so you don't have to remember to prime the coroutine).
    TODO: complete example
    '''
    def _impl(*args, **kwargs):
        routine = func(*args, **kwargs)
        routine.send(None)
        return routine
    return _impl


def iteratee_to_coroutine(func: Callable) -> Callable:
    '''
    '''
    @wraps(func)
    def _impl(target):
        while True:
            target.send(func((yield)))
    return _impl


@prime_coroutine
def broadcast(*targets: Coroutine) -> Generator:
    '''
    Send an item to multiple target coroutines.
    TODO: complete example
    '''
    while True:
        item = (yield)
        for target in targets:
            target.send(item)


@prime_coroutine
def printer():
    '''
    Sink for printing pipeline items.
    TODO: complete example
    '''
    while True:
        item = (yield)
        print(item)


@prime_coroutine
def pprinter():
    '''
    Sink for pretty printing pipeline items.
    TODO: complete example
    '''
    while True:
        item = (yield)
        pprint(item)
        print()
